#include <omp.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#define NUM_RUNS 1
#define NUM_INCS 10000
#define MAX_THREADS 2

typedef volatile struct _mcs_qnode_t {
  char flag;
  volatile struct _mcs_qnode_t* next;
} mcs_qnode_t;

void print_stats(char* lockname, int numthreads, double usecs) {
    static char first = 1;
    if (first) {
        printf("%10s %8s %8s\n", "lockname", "threads", "usecs");
        first = 0;
    }
    printf("%10s %8i %8.4lf\n", lockname, numthreads, usecs);
}

double diff(struct timespec start, struct timespec end) {

    struct timespec temp;
    double res;

    if ((end.tv_nsec-start.tv_nsec)<0) {
        temp.tv_sec = end.tv_sec-start.tv_sec-1;
        temp.tv_nsec = 1e9+end.tv_nsec-start.tv_nsec;
    } else {
        temp.tv_sec = end.tv_sec-start.tv_sec;
        temp.tv_nsec = end.tv_nsec-start.tv_nsec;
    }
    res = ((double) temp.tv_sec) * 1e6 + ((double) temp.tv_nsec) / 1e3;

    return res;

}

static inline unsigned long swap(volatile unsigned long* ptr, unsigned long val) {
    asm volatile("lock;"
                 "xchgq %0, %1"
                 : "=r"(val), "=m"(*ptr)
                 : "0"(val), "m"(*ptr)
                 : "memory");
    return val;
}

static inline unsigned long cas(volatile unsigned long* ptr, unsigned long old, unsigned long _new) {
    unsigned long prev;
    asm volatile("lock;"
                 "cmpxchgq %1, %2;"
                 : "=a"(prev)
                 : "r"(_new), "m"(*ptr), "a"(old)
                 : "memory");
    return prev;
}

static inline int bool_cas(volatile unsigned long* ptr, unsigned long old, unsigned long _new) {
    if (cas(ptr, old, _new) == old) return 1;
    return 0;
}

int benchmark_mcs_lock() {

   struct timespec time1, time2;
   volatile int* a = malloc(sizeof(int));
   volatile mcs_qnode_t* L = malloc(sizeof(mcs_qnode_t)); 
   L->flag=0;
   L->next = NULL;

    for (int threads=1; threads<=MAX_THREADS; threads++) {
        for (int run=0; run<NUM_RUNS; run++) {

            *a = 0;

            clock_gettime(CLOCK_REALTIME, &time1);
            #pragma omp parallel shared(L,a) num_threads(threads)
            {
                volatile mcs_qnode_t* I = malloc(sizeof(mcs_qnode_t));
                
                for (int i=0; i<NUM_INCS; i++) {
                    // lock
		            I->next = 0;
                    mcs_qnode_t* pred = (mcs_qnode_t*) swap((volatile unsigned long*) L, (unsigned long)I);

                    if (pred != 0) {
                        I->flag = 1;
                        pred->next = I;
                        while (I->flag) { } // spin
                    }

                    *a += 1;

                    // unlock
                    if (I->next == 0) {
                        if (bool_cas((volatile unsigned long*)L, (unsigned long)I, 0)) {
			                continue;
			            };
                        while (I->next == 0) { } // spin
                    }
                    I->next->flag = 0;
                    
                }

                free((void*)I);

            }

            clock_gettime(CLOCK_REALTIME, &time2);
            print_stats("MCS", threads, diff(time1, time2)/NUM_INCS);

            if (*a != threads*NUM_INCS) {
                fprintf(stderr, "Wrong result! [%i]\n", *a);
                exit(EXIT_FAILURE);
            }
        }

    }

    free((void*)a);

}

int main(int argc, char** argv) {

    benchmark_mcs_lock();

}



